#include "measure_backfacingness.h"
#include <vector>
#include <Eigen/Geometry>
#include <kt84/graphics/graphics_util.h>
#include <kt84/eigen_def.h>
#include <kt84/graphics/FramebufferObject.h>
#include <kt84/graphics/DisplayList.h>
#include <kt84/util.h>
#include <kt84/glut_util.h>
#define STB_IMAGE_WRITE_IMPLEMENTATION
#include "stb_image_write.h"

using namespace std;
using namespace Eigen;
using namespace kt84;
using namespace kt84::graphics_util;

double measure_backfacingness(const MatrixXd& V, const MatrixXi& F, int framebuffer_resolution, bool debug) {
    // OpenGL settings
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT, GL_DIFFUSE);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glEnable(GL_DEPTH_TEST);
    
    // Compute bounding box
    AlignedBox3d bbox;
    for (int i = 0; i < V.rows(); ++i)
        bbox.extend(Vector3d(V.row(i)));
    Vector3d bbox_diag   = bbox.diagonal();
    Vector3d bbox_center = bbox.center();
    const double R = bbox_diag.maxCoeff() * 1.05 / 2;
    
    // 6 views
    Vector3d camera_eye[6] = {
        bbox_center + 2 * Vector3d(R, 0, 0),
        bbox_center + 2 * Vector3d(0, R, 0),
        bbox_center + 2 * Vector3d(0, 0, R),
        bbox_center + 2 * Vector3d(-R, 0, 0),
        bbox_center + 2 * Vector3d(0, -R, 0),
        bbox_center + 2 * Vector3d(0, 0, -R),
    };
    Vector3d camera_up [6] = {
        Vector3d(0, 1, 0),
        Vector3d(1, 0, 0),
        Vector3d(0, 1, 0),
        Vector3d(0, 1, 0),
        Vector3d(1, 0, 0),
        Vector3d(0, 1, 0),
    };
    
    // Setup projection matrix
    glViewport(0, 0, framebuffer_resolution, framebuffer_resolution);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-R, R, -R, R, R, 3 * R);
    glMatrixMode(GL_MODELVIEW);
    
    // Create framebuffer
    TextureObject fbo_color;
    fbo_color.init();
    fbo_color.bind();
    fbo_color.allocate(framebuffer_resolution, framebuffer_resolution, GL_RGB, GL_RGB);
    RenderbufferObject fbo_depth;
    fbo_depth.init();
    fbo_depth.bind();
    fbo_depth.allocate(framebuffer_resolution, framebuffer_resolution, GL_DEPTH_COMPONENT16);
    FramebufferObject fbo;
    fbo.init();
    fbo.bind();
    fbo.attach_texture(GL_COLOR_ATTACHMENT0, fbo_color);
    fbo.attach_renderbuffer(GL_DEPTH_ATTACHMENT, fbo_depth);
    
    // Draw routine
    const Vector3uc color_front(255, 0, 0  );
    const Vector3uc color_back (  0, 0, 255);
    auto draw_mesh = [&](){
        glBegin(GL_TRIANGLES);
        for (int f = 0; f < F.rows(); ++f) {
            glColor3ub(color_front);
            for (int i = 0; i < 3; ++i) {
                int v = F(f, i);
                glVertex3d(V(v, 0), V(v, 1), V(v, 2));
            }
            glColor3ub(color_back);
            for (int i = 0; i < 3; ++i) {
                int v = F(f, 2 - i);
                glVertex3d(V(v, 0), V(v, 1), V(v, 2));
            }
        }
        glEnd();
    };
    DisplayList disp;
    
    // Render from 6 views
    int count_front = 0;
    int count_back  = 0;
    for (int i = 0; i < 6; ++i) {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        
        // Setup camera
        glLoadIdentity();
        gluLookAt(camera_eye[i], bbox_center, camera_up[i]);
        
        disp.render(draw_mesh);
        
        // Read back framebuffer
        vector<Vector3uc> pixels(framebuffer_resolution * framebuffer_resolution);
        fbo_color.copy_gpu2cpu(GL_UNSIGNED_BYTE, pixels[0].data());
        
        glutReportErrors();
        
        if (debug) {
            string filename = "debug_";
            util::append_str(filename, i);
            filename += ".png";
            if (stbi_write_png(filename.c_str(), framebuffer_resolution, framebuffer_resolution, 3, pixels[0].data(), 3 * framebuffer_resolution) == 0)
                cerr << "Failure while writing debug image!\n";
        }
        
        // Count pixels from front/back sides
#pragma omp parallel for reduction (+:count_front)
        for (int j = 0; j < pixels.size(); ++j)
            if (pixels[j] == color_front)
                ++count_front;
#pragma omp parallel for reduction (+:count_back)
        for (int j = 0; j < pixels.size(); ++j)
            if (pixels[j] == color_back)
                ++count_back;
    }
    
    return count_back / static_cast<double>(count_front + count_back);
}
