#include <iostream>
#include <sstream>
#include <fstream>
#include <boost/lexical_cast.hpp>
#include <igl/read.h>
#include <igl/write.h>
#include <igl/embree/reorient_facets_raycast.h>
#include <igl/unique_simplices.h>
using namespace std;
using namespace Eigen;

int main(int argc, char* argv[]) {
    auto print_usage = [&] () {
        cerr << "Usage: " << argv[0] << " [option(s)] input_mesh.{obj|off} output_mesh.{obj|off}\n"
            << "Options:\n"
            << "\t-t=<int> \t\t Total number of rays (default:10000)\n"
            << "\t-m=<int> \t\t Minimum number of rays for each patch (default:100)\n"
            << "\t-u \t\t\t Unify all duplicate facets into a unique one\n"
            << "\t-f \t\t\t Reorient face-wise (no manifold patches constructed)\n"
            << "\t-p \t\t\t Use parity mode\n"
            << "\t-z \t\t\t Skip writing output if no faces are flipped\n"
            << "\t-c \t\t\t Write as triangle soup with per-vertex color representing patch ID (only for *.obj)\n"
            << "\t-v \t\t\t Verbose messages in stdout\n"
            << endl;
        return 1;
    };

    if (argc < 3)
        return print_usage();

    // Parse command line input
    int rays_total = 10000;
    int rays_minimum = 100;
    bool unique_facet = false;
    bool facet_wise = false;
    bool use_parity = false;
    bool skip_zero = false;
    bool write_color = false;
    bool is_verbose = false;
    for (int i = 1; i < argc - 2; ++i) {
        string op = argv[i];
        string op_head = op.substr(0, 2);
        if (op_head == "-t") {
            if (op[2] != '=') {
                cerr << "Wrong format: " << op << endl;
                return print_usage();
            }
            rays_total = boost::lexical_cast<int>(op.substr(3, op.length() - 3));

        } else if (op_head == "-m") {
            if (op[2] != '=') {
                cerr << "Wrong format: " << op << endl;
                return print_usage();
            }
            rays_minimum = boost::lexical_cast<int>(op.substr(3, op.length() - 3));

        } else if (op_head == "-u") {
            unique_facet = true;

        } else if (op_head == "-f") {
            facet_wise = true;

        } else if (op_head == "-p") {
            use_parity = true;

        } else if (op_head == "-z") {
            skip_zero = true;

        } else if (op_head == "-c") {
            write_color = true;

        } else if (op_head == "-v") {
            is_verbose = true;

        } else {
            cerr << "Unrecognized option: " << op << endl;
            return print_usage();
        }
    }
    
    string input_filename  = argv[argc - 2];
    string output_filename = argv[argc - 1];

    if (write_color && output_filename.substr(output_filename.size() - 4, 4) != ".obj") {
        cerr << "-c option can only be used for writing mesh in *.obj format.\n";
        return 1;
    }
    
    // Read mesh
    if (is_verbose) cout << "Reading mesh: " << input_filename;
    MatrixXd V;
    MatrixXi F_in;
    if (!igl::read(input_filename, V, F_in)) {
        cerr << "Failed to read mesh!";
        return 1;
    }
    if (is_verbose) cout << " -> " << V.rows() << " vertives, " << F_in.rows() << " faces.\n";
    
    if (unique_facet) {
      if (is_verbose) cout << "Ensuring unique facets";
      MatrixXi F_unique;
      igl::unique_simplices(F_in, F_unique);
      if (F_unique.rows() < F_in.rows()) {
          if (is_verbose) cout << " -> reduced to " << F_unique.rows() << " faces.\n";
          F_in = F_unique;
      } else {
          if (is_verbose) cout << " -> all facets are already unique.\n";
      }
    }
    
    // Call igl::reorient_facets_raycast
    VectorXi I, C;
    igl::reorient_facets_raycast(V, F_in, rays_total, rays_minimum, facet_wise, use_parity, is_verbose, I, C);
    if (is_verbose) cout << I.sum() << " faces are flipped.\n";
    if (skip_zero && I.sum() == 0) {
        cout << "Skip writing output.\n";
        return 0;
    }
    
    // Output face indices by fliiping F_in according to I
    MatrixXi F_out = F_in;
    for (int f = 0; f < F_out.rows(); ++f)
        if (I[f] != 0) swap(F_out(f, 0), F_out(f, 2));
    
    // Write mesh
    if (is_verbose) cout << "Writing mesh: " << output_filename << endl;
    if (write_color) {
        // Partially copied from igl::writeOBJ
        ofstream s(output_filename.c_str());
        if (!s.is_open()) {
            cerr << "IOError: could not open " << output_filename << endl;
            return 1;
        }
        
        // Prepare vertex color table
        int num_patches = C.maxCoeff() + 1;
        MatrixXd color_table = 0.25 * (MatrixXd::Random(num_patches, 3) + MatrixXd::Constant(num_patches, 3, 3));
        
        // Write triangle soup
        for (int f = 0; f < (int)F_out.rows(); ++f) {
            Vector3d color = color_table.row(C(f));
            for (int i = 0; i < 3; ++i) {
                Vector3d pos = V.row(F_out(f, i));
                s << "v ";
                for (int j = 0; j < 3; ++j) s << pos  (j) << " ";
                for (int j = 0; j < 3; ++j) s << color(j) << " ";
                s << endl;
            }
        }
        for (int f = 0; f < (int)F_out.rows(); ++f) {
            s << "f " << (3 * f + 1) << " " << (3 * f + 2) << " " << (3 * f + 3) << endl;
        }
    } else if (!igl::write(output_filename, V, F_out)) {
        return 1;
    }
    
    return 0;
}
